#include <stdio.h>
#include <time.h>
#include <string.h>
#include <stdlib.h>
#include "rdtsc.h"
#include "test.h"

#include "inupfin.h"
/* Length of test block, number of test blocks.
 */

#define LENGTH_MESSAGE_TEST_ONE 63
#define LENGTH_MESSAGE_TEST_TWO 64 
#define LENGTH_MESSAGE_TEST_THREE 127
#define LENGTH_MESSAGE_TEST_FOUR  128
#define LENGTH_MESSAGE_TEST_FIVE 640
#define LENGTH_MESSAGE_TEST_SIX 128000
#define LENGTH_MESSAGE_TEST_SEVEN 1280
#define LENGTH_MESSAGE_TEST_EIGHT 12800
#define REPEAT_TEST_TIMES 100
#define REPEATHASH 100
#define NUMREPEATHASH 1000000
#define NUMTESTS 10


#define TEST_BLOCK_LEN 1024
#define TEST_BLOCK_COUNT 10240
#define BUFFER_SIZE 1024
#define HASH_BYTES HASH_BYTES_512

static void SHA3String(char *);
static void SHA3TimeTrial(void);
static void SHA3TestSuite(void);
static void SHA3File(char *);
static void SHA3Filter(void);
static void SHA3Print(unsigned char*);
static void SHA3BinaryTest(char *);
static void SHA3MultipleTimeTrial();
static void SHA3DanTimeTrial(unsigned int version);
static void MultipleUpdateSizeTest();
static void SHA3RDTSCTimeTrial(unsigned int version);
static void TestInitTime();

static unsigned int hashbitlen = 256;

/* Main driver.

 Arguments (may be any combination):
 -s<string> - digests string
 -t         - runs time trial
 -T 	    - another time trial
 -dan         another
  -r         -runs cycle test
 -x         - runs test script
 -u         - runs an update test, should fail the second and pass the others.
 -b<file>   - runs binary output test
 -<bitlen>  - sets the hash bitlen
 filename - digests file
 (none)   - digests standard input
other arguments are selfexplanatory
 */
int main(int argc, char *argv[]) {
	int i;
	if (argc > 1)
		for (i = 1; i < argc; i++)
			if (argv[i][0] == '-' && argv[i][1] == 'b')
				SHA3BinaryTest(argv[i] + 2);
			else if (argv[i][0] == '-' && argv[i][1] == 's')
				SHA3String(argv[i] + 2);
			else if (strcmp(argv[i], "-t") == 0)
				SHA3TimeTrial();
			else if (strcmp(argv[i], "-T") == 0)
				SHA3MultipleTimeTrial();
			else if (strcmp(argv[i], "-x") == 0)
				SHA3TestSuite();
			else if (strcmp(argv[i], "-u") == 0)
				MultipleUpdateSizeTest();
			else if (strcmp(argv[i], "-256") == 0)
				hashbitlen = 256;
			else if (strcmp(argv[i], "-512") == 0)
				hashbitlen = 512;
			else if (strcmp(argv[i], "-224") == 0)
				hashbitlen = 224;
			else if (strcmp(argv[i], "-384") == 0)
				hashbitlen = 384;
			else if (strcmp(argv[i], "-dan") == 0)
				SHA3DanTimeTrial(hashbitlen);
			else if (strcmp(argv[i], "-r") == 0)
				SHA3RDTSCTimeTrial(hashbitlen);
			else if (strcmp (argv[i], "-gs") == 0) /*greedysearchtests should only be used with 256*/
			   GreedySearch();
			else if (strcmp (argv[i], "-gs2") == 0)
			   GreedySearch2();
			else if (strcmp (argv[i], "-gs3") == 0)
			   GreedySearch3();
			else if (strcmp (argv[i], "-gs512") == 0)/*this one and next for the 512 version*/
			   GreedySearch512();
			else if (strcmp (argv[i], "-gs512double") == 0)
				GreedySearch512Double();
			else
				SHA3File(argv[i]);
	else
		SHA3TimeTrial();

	return (0);
}

/* Digests a string and prints the result.
 */
static void SHA3String(char *string) {
	hashState context;
	unsigned char digest[HASH_BYTES];
	size_t len = strlen(string);
	if (Init(&context, hashbitlen) != SUCCESS) {
		fprintf(stderr, "SHA3STring: BAD_HASHLEN\n");
		return;
	}
	Update(&context, (unsigned char*) string, len * 8);
	Final(&context, digest);
	printf("SHA-3 (\"%s\") = ", string);
	SHA3Print(digest);
	printf("\n");
}

char * string2binary(unsigned char* sequence, char* string, int bitlen) {
	int chars = bitlen >> 3;
	int i, j, c, bits;
	for (i = 0; i < chars; i++) {
		c = sequence[i];
		for (j = 7; j >= 0; j--) {
			string[i * 8 + j] = c & 1 ? '1' : '0';
			c >>= 1;
		}
	}
	bits = bitlen & 7;
	c = sequence[chars];
	for (i = 0; i < bits; i++) {
		string[chars * 8 + i] = c & (1 << (7 - i)) ? '1' : '0';
	}
	string[bitlen] = 0;
	return string;
}

/* Digests a bit pattern and prints the result.
 */
static void SHA3BitPattern(unsigned char sequence[], int bitlen) {
	hashState context;
	char* string;
	unsigned char digest[HASH_BYTES];
	if (Init(&context, hashbitlen) != SUCCESS) {
		fprintf(stderr, "SHA3BitPattern: BAD_HASHLEN\n");
		return;
	}
	Update(&context, sequence, bitlen);
	Final(&context, digest);
	string = malloc(bitlen + 1);
	printf("SHA-3 (%s) = ", string2binary(sequence, string, bitlen));
	free(string);
	SHA3Print(digest);
	printf("\n");
}

static void SHA3RepeatedPattern(unsigned char pattern, int count){
  BitSequence* sequence;
  hashState context;
  unsigned char digest[HASH_BYTES];
  if (Init (&context, hashbitlen) != SUCCESS){
  	fprintf(stderr, "SHA3RepeatedPattern: BAD_HASHLEN\n");
  	return;
  }
  sequence = malloc(count);
  memset(sequence,pattern,count);
  Update (&context, sequence, count*8);
  Final (&context, digest);
  free(sequence);
  printf ("SHA-3 (%d x 0x%02x) = ", count, pattern);
  SHA3Print (digest);
  printf ("\n");
}



/* Measures the time to digest TEST_BLOCK_COUNT TEST_BLOCK_LEN-byte
 blocks.
 */
static void SHA3TimeTrial() {
	hashState context;
	clock_t t1, t2;
	float seconds;
	static unsigned char block[TEST_BLOCK_LEN], digest[HASH_BYTES];
	unsigned int i;
	memset(block, 0, TEST_BLOCK_LEN);
	printf("SHA-3 time trial. Digesting %d %d-byte blocks ...",
			TEST_BLOCK_COUNT, TEST_BLOCK_LEN);
	fflush(stdout);
	/* Start timer */
	t1 = clock();
	/* Digest blocks */
	Init(&context, hashbitlen);
	for (i = 0; i < TEST_BLOCK_COUNT; i++)
		Update(&context, block, TEST_BLOCK_LEN*8);
	Final(&context, digest);

	/* Stop timer */
	t2 = clock();
	printf(" done\n");
	printf("Digest = ");
	SHA3Print(digest);
	seconds = (t2 - t1) / (float) CLOCKS_PER_SEC;
	printf("\nClock = %.4f seconds\n", seconds);
	printf("Speed = %.2f MB/second\n", (float) TEST_BLOCK_LEN
			* TEST_BLOCK_COUNT / seconds / 1024 / 1024);
}


/* Digests a reference suite of strings and prints the results.
 */

static void SHA3BinaryTest(char * filename){
	unsigned char digest[HASH_BYTES];
	int i;
	FILE *file;
	memset(digest,0,HASH_BYTES);
	if ((file = fopen(filename, "wb")) == NULL){
		printf("'%s' can't be opened\n", filename);
		return;
	}
	for (i = 0; i < 200000; i++){
		Hash(hashbitlen, digest, hashbitlen, digest);
		fwrite(digest, 1, hashbitlen/8, file);
	}
	fclose(file);
}

static void SHA3TestSuite() {
	static unsigned char ONES[] = {0xFF, 0xFF };
	static unsigned char ZEROS[] = {0x0, 0x0 };
	static unsigned char ZEROSANDONES[] = {0xAA, 0xAA };

	printf("SHA-3 test suite:\n");
	SHA3String("a");
	SHA3String("");
	SHA3String("1");
	SHA3String("a");
	SHA3String("1");
	SHA3String("2");
	SHA3String("b");
	SHA3String("c");
	SHA3String("abc");
	SHA3String("bcd");
	SHA3String("message digest");
	SHA3String("abcdefghijklmnopqrstuvwxyz");
	SHA3String("ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789");
	SHA3String(
			"1234567890123456789012345678901234567890\
1234567890123456789012345678901234567890");
	printf ("Bit patterns:\n");
	 SHA3BitPattern(ONES,0);
	 SHA3BitPattern(ONES,1);
	 SHA3BitPattern(ONES,3);
	 SHA3BitPattern(ONES,9);
	 SHA3BitPattern(ZEROS,1);
	 SHA3BitPattern(ZEROS,3);
	 SHA3BitPattern(ZEROS,9);
	 SHA3BitPattern(ZEROSANDONES,3);
	 SHA3BitPattern(ZEROSANDONES,9);
	 SHA3RepeatedPattern(0x0, 64);
	 SHA3RepeatedPattern(0xaa, 65);
	 }

/* Digests a file and prints the result.

 */
static void SHA3File(filename)
	char *filename; {
	FILE *file;
	hashState context;
	DataLength len, bytes_processed;
	double elapsedClock;
	unsigned char buffer[BUFFER_SIZE], digest[HASH_BYTES];
	clock_t startClock, endClock;
	bytes_processed = 0;

	if ((file = fopen(filename, "rb")) == NULL)
		printf("%s can't be opened\n", filename);

	else {
		startClock = clock();
		Init(&context, hashbitlen);
		len = (DataLength) fread(buffer, 1, BUFFER_SIZE, file);
		while (len) {
			bytes_processed += len;
			Update(&context, buffer, len * 8);
			len
					= (DataLength) fread(buffer, 1, BUFFER_SIZE,
							file);
		}
		Final(&context, digest);
		endClock = clock();
		fclose(file);

		printf("SHA-3 (%s) = ", filename);
		SHA3Print(digest);
		printf("\n");
		elapsedClock = (endClock - startClock) / (double) CLOCKS_PER_SEC;
		printf("Clock: %.2f seconds (%.2f MB/sec)\n bytes processed:%llu\n",
				elapsedClock, (double) bytes_processed / 1024 / 1024
						/ elapsedClock, bytes_processed);
	}
}

/* Digests the standard input and prints the result.
 */
static void SHA3Filter() {
	hashState context;
	size_t len;
	unsigned char buffer[BUFFER_SIZE], digest[HASH_BYTES];

	Init(&context, hashbitlen);
	while ((len = fread(buffer, 1, BUFFER_SIZE,stdin)))
		Update(&context, buffer, len * 8);
	Final(&context, digest);

	SHA3Print(digest);
	printf("\n");
}

/* Prints a message digest in hexadecimal.
 */
static void SHA3Print(unsigned char digest[]) {

	unsigned int i;
	for (i = 0; i < hashbitlen / 8; i++)
		printf("%02x", digest[i]);
}

static void TestUpdate(unsigned char *block, int len, int count, unsigned char* digest) {
	hashState context;
	int i;
	Init(&context, hashbitlen);
	for (i = 0; i < count; i++)
		Update(&context, block, len*8);
	Final(&context, digest);
}

static int equals(unsigned char* a, unsigned char* b){
	unsigned int i;
	for(i=0; i<hashbitlen/8; i++)
		if (a[i] != b[i]) return 0;
	return 1;
}


static void MultipleUpdateSizeTest(){
	unsigned char buffer[93555];
	unsigned char reference[HASH_BYTES], digest[HASH_BYTES];
	int total = 479001600;
	int first = sizeof(buffer);
	int sizes[] = { 1, 13, 2, 3, 5, 7, 8, 11, 15, 16, 25, 32, 63, 64, 128, 512, 1024, 25600,  0 }  ;
	int i;
	memset(buffer,0,sizeof(buffer));
	printf("Testing multiple update sizes. Hashing %i zeros...", total); 
	TestUpdate(buffer,first,total/first, reference);
	printf("Hash: ");
	SHA3Print(reference);
	printf("\n");
	for (i=0; sizes[i]; i++){
		printf("Calls: %i, length: %i: ", total/sizes[i], sizes[i]);
		TestUpdate(buffer,sizes[i], total/sizes[i], digest);
		printf("%s\n", equals(digest,reference) ? "OK" : "Error");
	}
}



static void OneLenTimeTrial(BitSequence* data, DataLength len, int times, double *speed) {
	uint32_t t1, t2;
	static unsigned char digest[HASH_BYTES];
	/* Start timer */
	t1 = HiResTime();
		Hash(hashbitlen, data,len*8, digest);
	t2 = HiResTime();
	*speed = (t2 - t1 - getdtMin())/(double) len;
}

static void SHA3MultipleTimeTrial(){
	static int header = 1;
	int sizes[] = { 16, 64, 128, 256, 512, 1024, 8192, 65536, 524288,0 };
	BitSequence* data;
	int i;
	int maxlen = 1024*1024*300;
	double speed;
	initdtMin();
	if (header) {
		printf("Len ");
		for(i=0; sizes[i]; i++) printf("%7d", sizes[i]);
		printf("\n");
		header = 0;
	}
	printf("%i ", hashbitlen);
	for(i=0; sizes[i]; i++){
		double min = 1e6;
		int j;
		data = calloc(1,sizes[i]);
		for (j= 0; j < 100; j++){
			OneLenTimeTrial(data, sizes[i],maxlen/sizes[i] , &speed);
			if (speed < min) min = speed;
		}
		free(data);
		printf("%7.2f", min);
	}
	printf("\n");
}


static void SHA3DanTimeTrial(unsigned int version) {
	/*hashState context;*/
	clock_t t1, t2;
	float seconds[NUMTESTS][REPEAT_TEST_TIMES];
	static unsigned char message[LENGTH_MESSAGE_TEST_SIX], digest[HASH_BYTES];
	static unsigned char message2[LENGTH_MESSAGE_TEST_SIX];
	static unsigned char message3[LENGTH_MESSAGE_TEST_SIX];
	static unsigned char message4[LENGTH_MESSAGE_TEST_SIX];
	unsigned int semilla;
	int i,j,k,numberrepeats;
	long long lentest[NUMTESTS];
	lentest[8]=LENGTH_MESSAGE_TEST_EIGHT<<3;
	lentest[1]=LENGTH_MESSAGE_TEST_ONE<<3;
	lentest[2]=LENGTH_MESSAGE_TEST_TWO<<3;
	lentest[4]=LENGTH_MESSAGE_TEST_THREE<<3;
	lentest[5]=LENGTH_MESSAGE_TEST_FOUR<<3;
	lentest[6]=LENGTH_MESSAGE_TEST_FIVE<<3;
	lentest[9]=LENGTH_MESSAGE_TEST_SIX<<3;
	lentest[7]=LENGTH_MESSAGE_TEST_SEVEN<<3;
	lentest[0]=LENGTH_MESSAGE_TEST_TWO<<2;
	lentest[3]=(LENGTH_MESSAGE_TEST_TWO<<3)+1;
	/*memset(block, 0, TEST_BLOCK_LEN);*/
	printf(" Dan time trial");
	fflush(stdout);
	  semilla=time(NULL);
	    /* Initialize block */
  for (i = 0; i <LENGTH_MESSAGE_TEST_SIX; i++){
        semilla=(semilla^i);
		semilla=semilla*((semilla<<1)+1);
		semilla=(semilla>>24)|(semilla<<8);
 		message[i] = (unsigned char)(semilla & 0xff);
		semilla=semilla*((semilla<<1)+1);
		semilla=(semilla>>24)|(semilla<<8);
 		message2[i] = (unsigned char)(semilla & 0xff);
		semilla=semilla*((semilla<<1)+1);
		semilla=(semilla>>24)|(semilla<<8);
 		message3[i] = (unsigned char)(semilla & 0xff);
		semilla=semilla*((semilla<<1)+1);
		semilla=(semilla>>24)|(semilla<<8);
 		message4[i] = (unsigned char)(semilla & 0xff);
	  }/*end for i */
	/*start tests*/
  printf("Test on %i version\n",version);
	  for(k=0;k<NUMTESTS;k++){
		  numberrepeats=(1+(NUMREPEATHASH/lentest[k]))*REPEATHASH;
		  printf("\n\n Test %i.\n Length of message:%llu bits.\n Time tested to hash %i such messages.\n Repeated %i times\n",k,lentest[k],numberrepeats,REPEAT_TEST_TIMES);
  for (i = 0; i < REPEAT_TEST_TIMES; i++){
	/* Start timer */
	
	t1 = clock();
	/* Digest  */
	for(j=0;j<numberrepeats;j++){
	Hash(hashbitlen,message,lentest[k],digest);
	Hash(hashbitlen,message2,lentest[k],digest);
	Hash(hashbitlen,message3,lentest[k],digest);
	Hash(hashbitlen,message4,lentest[k],digest);
/*	Init(&context, hashbitlen);
	Update(&context, block, 8);
	for (i = 0; i < TEST_BLOCK_COUNT; i++)
		Update(&context, block, TEST_BLOCK_LEN*8);
	Final(&context, digest);*/

	/* Stop timer */
	/*message=(message+j)&0x1ff;*/
	t2 = clock();
	}/*end for j*/
	seconds[k][i] = (t2 - t1) / (float) CLOCKS_PER_SEC;
	printf("%.4f seconds", seconds[k][i]);
	printf("(%.2f MB/second)", (float) numberrepeats*lentest[k]/2 / seconds[k][i] / 1024 / 1024);
	if(i%20==19) printf("\n");
  }/*end for i*/
	  }/*end for k*/
	printf(" done\n");
	printf("last Digest = ");
	SHA3Print(digest);

	
}

static void SHA3RDTSCTimeTrial(unsigned int version) {
	/*hashState context;*/
	uint32_t t1, t2;
	uint32_t seconds[NUMTESTS][REPEAT_TEST_TIMES];
	uint32_t min[NUMTESTS];
	static unsigned char message[LENGTH_MESSAGE_TEST_SIX], digest[HASH_BYTES];
	static unsigned char message2[LENGTH_MESSAGE_TEST_SIX];
	static unsigned char message3[LENGTH_MESSAGE_TEST_SIX];
	static unsigned char message4[LENGTH_MESSAGE_TEST_SIX];
	uint32_t dtMin;
	unsigned int semilla;
	int i,k;
	long long lentest[NUMTESTS];
	lentest[8]=LENGTH_MESSAGE_TEST_EIGHT<<3;
	lentest[1]=LENGTH_MESSAGE_TEST_ONE<<3;
	lentest[2]=LENGTH_MESSAGE_TEST_TWO<<3;
	lentest[4]=LENGTH_MESSAGE_TEST_THREE<<3;
	lentest[5]=LENGTH_MESSAGE_TEST_FOUR<<3;
	lentest[6]=LENGTH_MESSAGE_TEST_FIVE<<3;
	lentest[9]=LENGTH_MESSAGE_TEST_SIX<<3;
	lentest[7]=LENGTH_MESSAGE_TEST_SEVEN<<3;
	lentest[0]=LENGTH_MESSAGE_TEST_TWO<<2;
	lentest[3]=(LENGTH_MESSAGE_TEST_TWO<<3)+1;
	/*memset(block, 0, TEST_BLOCK_LEN);*/
	fflush(stdout);
	semilla=time(NULL);
	/* Initialize block */
	for (i = 0; i <LENGTH_MESSAGE_TEST_SIX; i++){
		semilla=(semilla^i);
		semilla=semilla*((semilla<<1)+1);
		semilla=(semilla>>24)|(semilla<<8);
		message[i] = (unsigned char)(semilla & 0xff);
	}/*end for i */
	/*start tests*/
	printf("Cycle Test on %i version\n",version);
	dtMin = initdtMin();
	for(k=0;k<NUMTESTS;k++)
		if(k!=4) printf("%7d ", lentest[k]);
	printf("\n");
	for(k=0;k<NUMTESTS;k++){
		if (k!=4){/*skip test 4 for formatting reasons*/
		min[k] = 0xFFFFFFFF;
		//numberrepeats=(1+(NUMREPEATHASH/lentest[k]))*REPEATHASH;
		//printf("\n\n Test %i.\n Length of message:%llu bits.\n  Repeated %i times\n",k,lentest[k],REPEAT_TEST_TIMES);
		for (i = 0; i < REPEAT_TEST_TIMES; i++){
			/* Start timer */

			t1 = HiResTime();
			/* Digest  */
			Hash(hashbitlen,message,lentest[k],digest);
			t2 = HiResTime() - t1 - dtMin;
			if (t2 < min[k]) min[k] = t2;
			seconds[k][i] = (t2);
			//printf("%.4f seconds", seconds[k][i]);
			//printf("(%.2f MB/second)", (float) numberrepeats*lentest[k]/2 / seconds[k][i] / 1024 / 1024);
			//if(i%20==19) printf("\n");
		}/*end for i*/
		printf("%7.2f ", min[k]/((float) lentest[k]));
		}}/*end for k*/
	for (i = 13; i < REPEAT_TEST_TIMES; i=i+11){
		printf("\n");
		for(k=0;k<NUMTESTS;k++) if(k!=4) printf("%7.2f ", seconds[k][i]/((float) lentest[k]));
	}printf("\n");
	for(k=0;k<NUMTESTS;k++) if(k!=4) printf("%7.2i ", min[k]);
	printf("\ndone\n");



}

