#include <string.h>
#include <stdio.h>
#include "inupfin.h"

#define STATE_DWORDS STATE_DWORDS_256
#define STATE_BYTES STATE_BYTES_256
#define BLOCK_DWORDS BLOCK_DWORDS_256
#define BLOCK_BYTES BLOCK_BYTES_256
#define HASH_BYTES HASH_BYTES_256
#define BLOCK_BITS 512

#define UPDATEBITS   state->bits_processed += BLOCK_BITS;

extern FILE *fp_out;

void Encrypt256(uint64_t hsh[STATE_DWORDS], uint64_t k[BLOCK_DWORDS],uint64_t pk[BLOCK_DWORDS],uint64_t bits);
void Transform256(hashState256* state);

const static unsigned char BITMASK[] = {0x00, 0x80, 0xc0, 0xe0, 0xf0, 0xf8, 0xfc, 0xfe};
const static unsigned char BITPADDING[] = { 0x80, 0x40, 0x20, 0x10, 0x08, 0x04, 0x02, 0x01 };

const uint64_t i256[STATE_DWORDS] =
{
    0x6a09e667f3bcc908ull, 0xbb67ae8584caa73bull,
    0x3c6ef372fe94f82bull, 0xa54ff53a5f1d36f1ull
};

const uint64_t  i224[STATE_DWORDS] =
{
    0x510e527fade682d1ull, 0x9b05688c2b3e6c1full,
    0x1f83d9abfb41bd6bull, 0x5be0cd19137e2179ull,
};




HashReturn Init256(hashState256 *state){
	state->previous_block = state->buffer;
	state->data_block = state->buffer+BLOCK_DWORDS;
	state->bits_processed = 0;
	state->bits_waiting_for_process = 0;
	memcpy(state->state, i256, STATE_BYTES);
	memcpy(state->previous_block, i224, STATE_BYTES);
	memcpy(state->previous_block+(BLOCK_DWORDS >> 1), i224, STATE_BYTES);
	return SUCCESS;
}

HashReturn Init224(hashState256 *state){
	state->previous_block = state->buffer;
	state->data_block = state->buffer+BLOCK_DWORDS;
	state->bits_processed = 0;
	state->bits_waiting_for_process = 0;
	memcpy(state->state, i224, STATE_BYTES);
	memcpy(state->previous_block, i256, STATE_BYTES);
	memcpy(state->previous_block+(BLOCK_DWORDS >> 1), i256, STATE_BYTES);
	return SUCCESS;
}



HashReturn Update256(hashState256 *state, const BitSequence *data,
		DataLength databitlen) {
	DataLength offset, databytes;
	unsigned int remaining, index;
	BitSequence* byte_buffer;
	databytes = (databitlen >> 3); /* number of bytes to process */
	index = (state->bits_waiting_for_process >> 3);/*bytes waiting to be processed
	 According to specifications, except for the last time, update will never be called with
	 a number of bits not multiple of 8. Therefore, the number of bits that have been
	 fed into the buffer but not processed after the last call of update must be
	 a multiple of 8. Also, this number should be less than BLOCK_BYTES*/
	remaining = BLOCK_BYTES - index;
	offset = 0;
	if ((index != 0) && (databytes >= remaining)) {
		byte_buffer = (BitSequence*) state->data_block; /* buffer as an array of char */
		memcpy(byte_buffer + index, data, remaining);
		UPDATEBITS;
		Transform256(state);
		offset += remaining;
		index = 0;
		state->bits_waiting_for_process = 0;
	}
	while (offset + BLOCK_BYTES <= databytes) {
		memcpy(state->data_block, data + offset, BLOCK_BYTES);
		UPDATEBITS;
		Transform256(state);
		offset += BLOCK_BYTES;
	}
	if (databitlen & 7)
		databytes++; /* The number of bits is not a multiple of 8. This can only happen in the last call to update */
	byte_buffer = (BitSequence*) state->data_block;
	memcpy(byte_buffer + index, data + offset, (size_t)(databytes - offset));
	state->bits_waiting_for_process += (uint32_t) (databitlen - 8 * offset);;
	return SUCCESS;
}


void Transform256(hashState256* state){
	uint64_t temp[STATE_DWORDS];
	uint64_t* swap_aux;
	int i;
	memcpy(temp,state->state,STATE_BYTES);
	Encrypt256(temp, state->data_block, state->previous_block,state->bits_processed);
	for (i = 0; i < STATE_DWORDS; i++)
		state->state[i] ^= temp[i];
	swap_aux = state->data_block;
	state->data_block = state->previous_block;
	state->previous_block = swap_aux; /* now previous_block points to the block just processed */
	fprintf(fp_out, "\n########################################################################################\n\n");
	fprintf(fp_out,"Davies-Meyer:0x%016llx 0x%016llx 0x%016llx 0x%016llx\n",state->state[0],state->state[1],state->state[2],state->state[3]);
	fprintf(fp_out, "\n########################################################################################\n\n");
}




HashReturn _Final256 (hashState256 *state, BitSequence *hashval, int hashbytes){
	int j;
	BitSequence* byte_buffer;
	unsigned int index, bits;
	if(state->bits_waiting_for_process){/*if there is any data to be processed,
				 	we need to padd and process, otherwise we
					go to the last block*/

		index = state->bits_waiting_for_process >> 3;
		bits =  state->bits_waiting_for_process & 7;
		byte_buffer = (BitSequence*) state->data_block;
		byte_buffer[index] = (byte_buffer[index] & BITMASK[bits]) | BITPADDING[bits];
		index++;
		memset(byte_buffer+index, 0, BLOCK_BYTES - index);
		state->bits_processed += state->bits_waiting_for_process; /*the usual update of bits processed cannot be used here, since
		less than a full block of DATA has been processed*/

		Transform256(state);
	}/*endif(bits_waiting_for_process)*/
   /*all data (plus any padding) has been processed. A last call, so that the last block
   of data is treated like all the others (i.e., twice) is made. The new block is simply the number of
   bits processed. */
	state->data_block[0]=state->bits_processed;
	for(j=1;j < STATE_DWORDS;j++) state->data_block[j]=0; /* The size of the block is twice the size of the state */
	memcpy(state->data_block + STATE_DWORDS, state->state, STATE_BYTES);
	state->bits_processed = 0; /* This ensures that the final transform is the same regardless of the bits processed */
	Transform256(state);
	for (j = 0; j < hashbytes; j++)
		hashval[j] = ((BitSequence*)state->state)[j];
	return SUCCESS;
}


HashReturn Final256(hashState256 *state, BitSequence *hashval){
	return _Final256(state,hashval,HASH_BYTES_256);
}

HashReturn Final224(hashState256 *state, BitSequence *hashval){
	return _Final256(state,hashval,HASH_BYTES_224);
}







HashReturn Hash256(const BitSequence *data, DataLength databitlen, BitSequence *hashval){
	hashState256 state;
	HashReturn status;
	status = Init256(&state);
	if (status != SUCCESS) return status;
	status = Update256(&state, data, databitlen);
	if (status != SUCCESS) return status;
	return _Final256(&state, hashval,HASH_BYTES_256);
}

/* Hash the supplied data and provide the resulting hash value. 224 bit version
 * Parameters:
 *	data: the data to be hashed
 *	databitlen: the length, in bits, of the data to be hashed
 *	hashval: the resulting hash value of the provided data
 * Returns:
 *	SUCCESS - on success
*/
HashReturn Hash224(const BitSequence *data, DataLength databitlen, BitSequence *hashval){
	hashState256 state;
	HashReturn status;
	status = Init224(&state);
	if (status != SUCCESS) return status;
	status = Update256(&state, data, databitlen);
	if (status != SUCCESS) return status;
	return _Final256(&state, hashval,HASH_BYTES_224);
}










